/*
 *  file: TABIFY converts spaces to tabs, but only where it will compact data.
 *
 */
#include	<alloc.h>
#include	<io.h>
#include	<ctype.h>
#include	<dir.h>
#include	<dos.h>
#include	<fcntl.h>
#include	<conio.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<sys\stat.h>
#include	<string.h>
#include	<ctype.h>


typedef	unsigned int	uint;
typedef	unsigned char	uchar;
typedef	unsigned long	ulong;
typedef	unsigned char	STREAM;


#define	TABWIDTH	8


/*
 *  PROTOTYPES for functions in this module.
 */
int	build_path(char *spec, char *drv, char *pth, char *fil, char *ext,
		char *prefix, char *complete, int *flag);
void	process_files(void);
int	insert_tabs(void);
int	getyn(void);
void	incr_time(void);
void	usage(void);
int	valid_text(void);


/*
 *  STORAGE
 */
STREAM	huge *fbuf;			/* file buffer			    */
STREAM	huge *iptr,  huge *optr;
long	buf_size;
long	isiz, osiz;
int	ihandle, ohandle;
char	ipath[MAXPATH], opath[MAXPATH];
char	iprefix[MAXPATH], oprefix[MAXPATH];
char	idrv[MAXDRIVE], odrv[MAXDRIVE];
char	idir[MAXDIR], odir[MAXDIR];
char	ifnm[MAXFILE], ofnm[MAXFILE];
char	iext[MAXEXT], oext[MAXEXT];
char	icomplete[MAXPATH], ocomplete[MAXPATH];
int	iflag, oflag;
int	blanks;
int	doquery;

/*
 *  MAIN program for tabifying text file.
 */
void
main(int argc, char **argv)
{
    int ival, argcount, n;
    char *args[10];

    _fmode = O_TEXT;
    printf(" Tabify: by M.Garvin\n");
    argcount = 0;
    doquery = 1;
    for (n = 1; n < argc; n++) {
	if (argv[n][0] == '-') {	/* option			    */
	    if (tolower(argv[n][1]) == 's')
		doquery = 0;
	} else {
	    args[argcount] = argv[n];
	    ++argcount;
	}
    }
    if (argcount != 1) {		/* only single file arg/wildcard    */
	usage();
	return;
    }
    ival = build_path(args[0],idrv,ipath,ifnm,iext,iprefix,icomplete,&iflag);
    if (!ival) {
	printf("Source path not valid.\n");
	return;
    }
    printf("\nProcessing file(s): %s\n", icomplete);
    /*
     *  Alloc a single buffer, as large as possible.
     */
    buf_size = ((unsigned long)coreleft() - 100L);
    fbuf = farmalloc(buf_size);
    if (fbuf == NULL) {
	printf("Allocation failed: Memory may be corrupted!\n");
	printf("Advise reboot and try again.\n");
	return;
    }
    process_files();
    farfree((char far *)fbuf);		/* free file buffer		    */
}


/*
 *  INSERT_TABS() converts appropriate blanks to tabs as it walks thru the
 *   file image.  The same buffer is used for input and output since the
 *   output file will always be equal or smaller than the input file.
 *
 *   On entry, it expects:
 *	fbuf has input file image.
 *	iptr points to the last byte in the input file image.
 *   On return:
 *	optr is left pointing 1 char past last char in output buffer.
 *	return 1 if file has been changed, else 0.
 */
int
insert_tabs(void)
{
    STREAM thischar;
    STREAM huge *tptr;
    long fsiz;
    int column, changed;

    fsiz = iptr - fbuf;
    column = 0;
    changed = 0;			/* false if no new tabs inserted    */
    tptr = fbuf;			/* point to text		    */
    optr = fbuf;			/* write back into the same buffer  */
    while (fsiz--) {
	thischar = *tptr++;		/* keep local for quick processing  */
	++column;
	switch (thischar) {
	case ' ':
	    if (column % TABWIDTH == 0)	{
		if (blanks) {		/* tab only if more than one space  */
		    *optr++ = '\t';
		    changed = 1;
		} else {
		    *optr++ = ' ';
		}
		blanks = 0;
	    } else {
		++blanks;
	    }
	    break;
	case '\n':
	case '\t':
	    *optr++  = thischar;
	    blanks = 0;
	    column = 0;
	    break;
	default:	
	    while (blanks > 0) {
		*optr++ = ' ';
		--blanks;
	    }
	    *optr++ = thischar;
	    break;
	}
    }
    return changed;
}


/*
 *  VALID_TEXT attempts to tell if valid text exists in file buffer.
 */
int
valid_text(void)
{
    STREAM huge *tptr;
    long fsiz;
    int maxline, thisline;

    fsiz = iptr - fbuf;
    tptr = fbuf;			/* point to text		    */
    maxline = 0;			/* keep count of longest line	    */
    thisline = 0;			/* count chars in current line	    */
    while (fsiz--) {
	if (*tptr == '\n') {
	    if (thisline > maxline)
		maxline = thisline;
	    thisline = 0;
	}
	++thisline;
	++tptr;
    }
    if (maxline > 160 || thisline > 160) /* check both in case \n never hit */
	return(0);
    return(1);
}


/*--------------------------- FILE FUNCTIONS -------------------------------*/


struct ftime itime;			/* gets time / date from file	    */


/*
 *  PROCESS_FILES reads in successive files matching passed file name.
 *    Processes the data and outputs to destination file.
 *    Returns when no more files match the passed name.
 */
void
process_files(void)
{
    int bytes, ilen, olen, badfile;
    char ifn[MAXPATH], ofn[MAXPATH];
    char *ext;
    struct ffblk fmainblock;		/* struct used for findfirst/next   */

    strcpy(ifn, iprefix);		/* copy drive, and path for source  */
    ilen = strlen(ifn);
    strcpy(ofn, oprefix);		/* copy drive, and path for dest    */
    olen = strlen(ofn);
    /*
     *  Find the first file.
     */
    if (findfirst(icomplete, &fmainblock, FA_ARCH)) {
	printf("File not found.\n");
	return;				/* not found			    */
    }
    /*
     *  Loop for all files matching the file spec.
     */
    do {
	/*
	 *  Check for .EXE, .ZIP, etc.
	 */
	ext = strrchr(fmainblock.ff_name, '.');
	if (ext != NULL) {		/* file extension exists	    */
	    badfile = 0;
	    if (!stricmp(ext + 1, "EXE"))
		badfile = 1;
	    else if (!stricmp(ext + 1, "COM"))
		badfile = 1;
	    else if (!stricmp(ext + 1, "ZIP"))
		badfile = 1;
	    else if (!stricmp(ext + 1, "OVL"))
		badfile = 1;
	    else if (!stricmp(ext + 1, "TAZ"))
		badfile = 1;
	    else if (!stricmp(ext + 1, "TAR"))
		badfile = 1;
	    else if (!stricmp(ext + 1, "OBJ"))
		badfile = 1;
	    else if (!stricmp(ext + 1, "LZH"))
		badfile = 1;
	    if (badfile) {
		printf("\nBad file extension in wildcard/filespec!\n");
		printf("Make sure filespec excludes EXE, COM, OBJ, etc.\n");
		exit(1);
	    }
	}
	/*
	 *  File exists:  find out if there is enough ram to load it.
	 */
	strcpy(&ifn[ilen], fmainblock.ff_name); /* join path + filename	    */
	isiz = fmainblock.ff_fsize;	/* length of file		    */
	if (isiz > buf_size) {
	    printf("Not enough memory to load: %s!\n", ifn);
	    printf(" -- Skipping this file.  Continue? (Y/N): ");
	    if (getyn() == 0)
		return;
	    if (findnext(&fmainblock))	/* find next file and continue	    */
		continue;
	    return;			/* no more files		    */
	}
	/*
	 *  Try to open the file and read it into the buffer.
	 */
	ihandle = open(ifn, O_RDONLY);
	if (ihandle == -1) {
	    printf("Error on open for: %s\n\n", ifn);
	    printf(" -- Skipping this file.  Continue? (Y/N): ");
	    if (getyn() == 0)
		return;
	    if (!findnext(&fmainblock))	/* find next file and continue	    */
		continue;
	    return;			/* no more files		    */
	}
	/*
	 *  Get time stamp for source file and increment it.
	 */
	getftime(ihandle, &itime);
	incr_time();
	/*
	 *  Read source file.
	 */
	iptr = fbuf;			/* fbuf remains stationary	    */
	printf("Reading: %s", ifn);
	do {				/* loop and read data		    */
	    bytes = read(ihandle, (char *)iptr, 32000);
	    iptr += bytes;
	} while (bytes);		/* quit when no bytes left	    */
	/*
	 *  Close input file and process data.
	 *  THIS IS WHERE OPERATION IS PERFORMED ON FILE IMAGE.
	 */
	close(ihandle);
	printf(" ...processing...\n");
	if (!valid_text()) {
	    printf("%s may not be a text file!\n", ifn);
	    printf(" Insert tabs anyway? (Y/N) 'Y' may corrupt this file!: ");
	    if (getyn() == 0) {
		if (!findnext(&fmainblock)) /* find next file and continue  */
		    continue;
		return;			/* no more files		    */
	    }
	}
	if (!insert_tabs()) {
	    if (!doquery)
		continue;
	    printf("No tabs needed: rewrite anyway? (Y/N) ");
	    if (!getyn())
		continue;		/* avoid writing output file	    */
	}
	/*
	 *  Open output file and write new data out to file.
	 */
	strcpy(&ofn[olen], fmainblock.ff_name); /* join path + filename	    */
	printf("       Writing: %s\n", ofn);
	ohandle = creat(ofn, S_IWRITE);
	if (ohandle == -1) {
	    printf("Cannot open output file: %s\n", ofn);
	    return;
	}
	osiz = optr - fbuf;
	optr = fbuf;
	do {
	    bytes = (int)min(osiz, 32000L); /* no of bytes to write	    */
	    write(ohandle, (char *)optr, (uint)bytes);
	    optr += bytes;
	    osiz -= bytes;
	} while (bytes > 0);
	/*
	 *  Set new time/date stamp for the output file before closing it.
	 */
	setftime(ohandle, &itime);
	close(ohandle);
	/*
	 *  Assume that the file spec was a wild card.  Try to find the next
	 *   file matching the file spec.
	 */
    } while (!findnext(&fmainblock));
    return;
}


/*
 *  BUILD_PATH constructs a complete description of path from source string.
 *   Pass in a pointer to the string to be parsed and pointers to the
 *   destinations for output data.
 *  Return 1 if spec is valid, else 0.
 */
int
build_path(char *spec, char *drv, char *pth, char *fnam, char *ext,
	char *prefix, char *complete, int *flag)
{
    int x;
    char temp[MAXDIR];
    struct ffblk fb;

    strupr(spec);			/* use upper case to be consistent  */
    *flag = fnsplit(spec, drv, pth, fnam, ext); /*split spec into components*/
    /*
     *  If the drive wasn't specified, find the current drive
     */
    if (!(*flag & DRIVE)) {
	strcpy(drv, "A:");
	*drv += getdisk();		/* Add drive no to 'A' to get ascii */
    }
    if (*flag & DIRECTORY) {
	/*
	 *  If directory spec'd, make sure '\' in front
	 */
	if (*pth != '\\') {
	    strcpy(temp, pth);		/* save a copy of relative dir given*/
	    *pth = '\\';		/* get dir relative to root	    */
	    getcurdir(*drv + 1 - 'A', pth + 1);
	    x = strlen(pth);
	    if (x > 1) {		/* if in root dir, avoid two \\'s   */
		pth[x] = '\\';
		strcpy(pth + x + 1, temp); /* copy primary dir name on top  */
	    } else
		strcpy(pth + 1, temp);	/* copy relative dir on top	    */
	}
    } else {				/* directory not spec'd		    */
	/*
	 *  If dir NOT spec'd, then get the name of the current directory
	 */
	*pth = '\\';
	getcurdir(*drv + 1 - 'A', pth+1);
	x = strlen(pth);
	if (x > 1)
	    pth[x] = '\\';
	pth[x+1] = 0;
    }
    /*
     *  At this point, the directory has been fully specified relative to root
     *  To find out if the parsed file name is actually a directory name,
     *  try to open *spec as directory. If so, then tack it onto the end of
     *  the subdir name and use '*.*' as the file designator.
     */
    if (*flag & FILENAME && !(*flag & WILDCARDS)) { /*specific filename?    */
	if (!findfirst(spec, &fb, FA_DIREC) && fb.ff_attrib & FA_DIREC) {
	    /*
	     *  File name given is actually a subdirectory name
	     */
	    strcat(pth, fnam);		/* add name to end of dir name	    */
	    strcat(pth, "\\");		/* tack a '\' on the end of filename*/
	    if (*flag & EXTENSION)	/* and extension, if it exists	    */
		strcat(pth, ext);	/* usually don't see subdirs w .EXT!*/
	    strcpy(fnam, "*");		/* and replace file name w wildcard */
	    *flag |= WILDCARDS;		/* reflect changes in flags	    */
	    *flag &= ~FILENAME;
	    *flag &= ~EXTENSION;
	}
    }
    /*
     *  No extension given or no file name given -- use wildcard
     */
    if (!(*flag & EXTENSION)) {
	strcpy(ext, ".*");
	*flag |= WILDCARDS;
    }
    if (!(*flag & FILENAME)) {
	strcpy(fnam, "*");
	*flag |= WILDCARDS;
    }
    fnmerge(complete, drv, pth, fnam, ext);
    fnmerge(prefix, drv, pth, "", "");	/* save prefix w/o filename	    */
    strcpy(temp, prefix);		/* derive prefix without '\'	    */
    x = strlen(temp);
    if ((*flag & DIRECTORY) && x > 1) {
	temp[x-1] = 0;
	if (!findfirst(temp, &fb, FA_DIREC) && fb.ff_attrib & FA_DIREC)
	    return(1);
	else
	    return(0);
    } else {
	return(1);			/* reference to root directory	    */
    }
}


void
incr_time()
{
    uint sec, min, hour, day, mon, year;

    sec  = itime.ft_tsec * 2;
    min  = itime.ft_min;
    hour = itime.ft_hour;
    day  = itime.ft_day;
    mon  = itime.ft_month;
    year = itime.ft_year;
    if ((sec += 2) >= 60) {
	sec = 0;
	if (++min >= 60) {
	    min = 0;
	    if (++hour >= 24) {
		hour = 1;
		if (++day >= 28) {
		    day = 1;
		    if (++mon >= 12) {
			mon = 1;
			++year;
		    }
		}
	    }
	}
    }
    itime.ft_tsec	= sec / 2;
    itime.ft_min	= min;
    itime.ft_hour	= hour;
    itime.ft_day	= day;
    itime.ft_month	= mon;
    itime.ft_year	= year;
}


/*
 *  GETYN loops until it sees a Y or N key (or ESC).  Returns 1 or 0 resp.
 */
int
getyn()
{
    char gc;

    while (1) {
	gc = toupper(getch());
	if (gc == 3) {
	    printf("\nAborting operation\n");
	    exit(1);
	}
	if (gc == 'Y' || gc == 'N' || gc == 27) {
	    printf("%c\n", gc);
	    return(gc == 'Y');		/* return 1 only for 'Y'	    */
	}
    }
}


char *utext[] = {
   " Usage: TABIFY FILENAME\n",
   "  Replaces spaces with tabs, but only where it will compact the file.\n",
   "  The maximum file size is limited by the amount of available memory.\n",
   "  This program attempts to increment the file's date/time stamp by 1\n",
   "  second, so the new file will be backed up by programs that check\n",
   "  time stamps.  However the 'vintage' of the file remains unaltered.\n",
   "  Since tabify will accept wildcards this version attempts to detect\n",
   "  whether a file contains valid text, in order to avoid overwriting\n",
   "  non-text files.  If the file does not look like a text file, tabify\n",
   "  will inform the user and offer to abort.\n"
};


void
usage()
{
    int lines, t;

    lines = sizeof(utext) / sizeof(char *);
    for (t = 0; t < lines; t++)
	printf(utext[t]);
}

